/*
 * Copyright (C) 2002-2006 StarBED Project All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * ccftp - concurrent ftp client
 *
 * $Id: ccftp.c,v 1.6 2007/01/23 04:14:32 k-chinen Exp $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <pthread.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <syslog.h>

#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#ifdef HAVE_POLL_H
#include <poll.h>
#endif

#include <sys/wait.h>
#include <sys/resource.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <errno.h>


#include "alist.h"
#include "msg.h"
#include "socket.h"
#include "sman.h"
#include "sigs.h"

#define SUB_MSEC(x,y) \
        ( (x.tv_sec  - y.tv_sec )*1000 + (x.tv_usec - y.tv_usec)/1000 )

#define Trace   RawTrace

#ifndef NTH_MAX
#define NTH_MAX     (129)
#endif
#ifndef PG_TIMEOUT
#define PG_TIMEOUT  (1201)
#endif


extern volatile int nret_http_read_timeout;
int download(char *src, char *dst, off_t*);
int upload(char *src, char *dst);


#ifdef HAVE_NANOSLEEP
int
msleep(int x)
{
    struct timespec w;
    long long tw;

    /* convert msec to nsec */
    tw = (long long)x * 1000000;
    w.tv_nsec = tw % 1000000000;
    w.tv_sec =  tw / 1000000000;

    nanosleep(&w, NULL);

    return 0;
}
#else
int
msleep(int x)
{
    int chk;
    struct timeval tv;

    tv.tv_usec = x % 1000;
    tv.tv_sec  = x / 1000;
try_select:
    chk = select(0, NULL, NULL, NULL, &tv);
    if(chk<0) {
        fprintf(stderr,
            "msleep[select]: fail select() (%d)\n", errno);
    }
    return 0;
}
#endif


#define _KB_CONST_  (1024L)
#define _MB_CONST_  (1024L*1024L)
#define _GB_CONST_  (1024L*1024L*1024L)

char*
r2h(char *dst, double t, int tobit)
{
    int f;
    double v;

    if(t<0) {
        f = -1;
        t = -t;
    }
    else {
        f = 1;
    }

    if(tobit) {
        t = t*8.0;
    }

    if(t>(double)_GB_CONST_) {
        v = t/((double)_GB_CONST_);
        if(f>0)
            sprintf(dst, "%7.2fG", v);
        else
            sprintf(dst, "%7.2fG", -v);
    }
    else {
        if(t>(off_t)_MB_CONST_) {
            v = t/((double)_MB_CONST_);
            if(f>0)
                sprintf(dst, "%7.2fM", v);
            else
                sprintf(dst, "%7.2fM", -v);
        }
        else {
            if(t>(off_t)_KB_CONST_) {
                v = t/((double)_KB_CONST_);
                if(f>0)
                    sprintf(dst, "%7.2fK", v);
                else
                    sprintf(dst, "%7.2fK", -v);
            }
            else {
                sprintf(dst, "%7.2f", t);
            }
        }
    }

    if(tobit) {
        strcat(dst, "bps");
    }
    else {
        strcat(dst, "B/s");
    }

    return dst;
}



/* the number of threads */
int nth_expect=1;
char *target_url=NULL;
int pg_timeout=PG_TIMEOUT;
int do_writefile=0;
char *outpathmask="d%d-%d.out";

int iter_expect=1;

static pthread_mutex_t slock = PTHREAD_MUTEX_INITIALIZER;
static int nth_up=0;
static int nth_fail=0;
static int nss_fine=0;
static int nss_err=0;
static int nss_mid_expect=0;

static pthread_mutex_t gate = PTHREAD_MUTEX_INITIALIZER;

int    mon_interval = 3;




void*
runner(void*ag)
{
    int runner_id;
    int i;
    char tmp[BUFSIZ];
    char msg[BUFSIZ];
    int ck;
    int mdiff;
    struct timeval mbegin, mend;
    off_t flen;
    double awr;
    double hawr;
    char hm[BUFSIZ];
    int icount;
    char *smsg;
#ifdef ERRTEST
    char dummy_url[BUFSIZ];
#endif


    runner_id = *(int*)ag;
    Trace("th# %3d: wait gate openning\n", runner_id);

    pthread_mutex_lock(&slock);
    nth_up++;
    pthread_mutex_unlock(&slock);

    /*
     * wait gete openning
     */
    pthread_mutex_lock(&gate);
    pthread_mutex_unlock(&gate);

    /*
     * main body
     */
    for(icount=0;icount<iter_expect;icount++) {
        if(do_writefile) {
            sprintf(tmp, outpathmask, runner_id, icount);
        }
        else {
            strcpy(tmp, "/dev/null");
        }

        gettimeofday(&mbegin, NULL);
#ifdef ERRTEST
        if(rand()%100<ERRFACTOR) {
            Trace("insert dummy URL in th# %3d\n", runner_id);
            strcpy(dummy_url, target_url);
            strcat(dummy_url, "X");
        }
        else {
            strcpy(dummy_url, target_url);
        }
        ck = download(dummy_url, tmp, &flen);
#else
        ck = download(target_url, tmp, &flen);
#endif
        gettimeofday(&mend, NULL);
        Trace("th# %3d; ck %d\n", runner_id, ck);
        if(ck) {
            smsg = "ERR";
        }
        else {
            smsg = "FINE";
        }

        mdiff = SUB_MSEC(mend, mbegin);
        if(mdiff<0) {
            awr = -1.0;
        }
        else {
            awr = ( (double)flen * 1000.0) / mdiff;
        }
        hawr = awr;
        r2h(hm, hawr, 1);
        sprintf(msg, "TH %d %d %d %s %12.2f B/s ; %lld B / %d ms %s\n",
            nth_expect, runner_id, icount,
            hm, awr, (long long)flen, mdiff, smsg);
        fputs(msg, stdout);
        fflush(stdout);
        fputs(msg, stderr);
        fflush(stderr);

        pthread_mutex_lock(&slock);
        if(ck) {
            nss_err++;
        }
        else {
            nss_fine++;
        }
        pthread_mutex_unlock(&slock);
    }

    pthread_mutex_lock(&slock);
    nth_up--;
    pthread_mutex_unlock(&slock);

    pthread_exit(NULL);
}

int
usage(char* pn)
{
    printf("ccftp - concurrent FTP client $Revision: 1.6 $\n");
#ifdef ERRTEST
    printf("*** compiled for ERRTEST with %d\n", ERRFACTOR);
#endif
    printf("usage:\n");
    printf("  %s [options] <URL>\n", pn);
    printf("options:\n");
    printf("  -h        print this message\n");
    printf("  -n num    the number of session(threads)     ; %d\n",
            nth_expect);
    printf("  -i num    the number of iteration            ; %d\n",
            iter_expect);
    printf("  -F        write downloaded files             ; ");
    if(do_writefile) {
        printf("ENABLE\n");
    }
    else {
        printf("DISABLE -> /dev/null\n");
    }
    printf("  -o path   path for download                  ; '%s'\n",
        outpathmask);
    printf("  -T num    session timeout                    ; %9d [s]\n",
        pg_timeout);
    printf("  -R num    HTTP read timeout                  ; %9d [us]\n",
        nret_http_read_timeout);
    printf("  -m num    monitoring interval                ; %9d [s]\n",
        mon_interval);

    printf("example:\n");
    printf("  %% ccftp -n 8 ftp://ftp.jaist.ac.jp/pub/FreeBSD/README.TXT\n");
    printf("  %% ccftp -n 4 -i 4 ftp://ftp.jaist.ac.jp/pub/FreeBSD/README.TXT\n");
    printf("  %% ccftp -n 3 -o '/tmp/x%%d-%%d' ftp://ftp.jaist.ac.jp/\n");
}


/* for nret.c */
extern pthread_mutex_t info_lock;
extern long long download_sum;
extern long long download_elen;

int
main(int argc, char *argv[])
{
    int flag;
    int i;
    int ck;
    pthread_t nn;
    int xtup, xtexp;
    int xserr, xsfine, xsexp;
    time_t tstart;
    time_t tnow;
    time_t tend;
    int    tdiff;
    long long glen;
    long long gsum;
    long long gesum;
    double gbprog;
    double gsprog;
    int rcount;

    
    while((flag=getopt(argc, argv, "hn:i:m:T:R:Fo:"))!=-1) {
        switch(flag) {
        case 'h':
            usage(argv[0]);
            exit(0);
            break;
        case 'n':
            nth_expect = atoi(optarg);
            break;
        case 'i':
            iter_expect = atoi(optarg);
            break;
        case 'm':
            mon_interval = atoi(optarg);
            break;
        case 'T':
            pg_timeout = atoi(optarg);
            break;
        case 'R':
            nret_http_read_timeout = atoi(optarg);
            break;
        case 'F':
            do_writefile = 1;
            break;
        case 'o':
            outpathmask = strdup(optarg);
            break;
        }
    }

    if(argv[optind]) {
        target_url = strdup(argv[optind]);
    }
    if(!target_url) {
        fprintf(stdout, "no target\n");
        usage(argv[0]);
        exit(77);
    }
    if(!*target_url) {
        fprintf(stdout, "empty target\n");
        usage(argv[0]);
        exit(97);
    }

    Trace("target '%s'\n", target_url);
    Trace("%d threads are expected\n", nth_expect);
    Trace("%d iterations are expected\n", iter_expect);
    Trace("timeout %ds\n", pg_timeout);

    pthread_mutex_lock(&gate);

    SetSignalHandler(SIGPIPE,   SIG_IGN);

    for(i=0;i<nth_expect;i++) {
        if(i>NTH_MAX) {
            break;
        }

        ck = pthread_create(&nn, NULL, runner, &i);

        msleep(3);
    }

    pthread_mutex_lock(&slock);
    xtup  = nth_up;
    xtexp = nth_expect;
    pthread_mutex_unlock(&slock);

    if(xtup<xtexp) {
        Error("sorry, too small thread (expected %d, upped %d)",
            xtexp, xtup);
        exit(43);
    }

    tstart = time(NULL);
    pthread_mutex_unlock(&gate);

    rcount = 0;
    while(1) {
        tnow = time(NULL);
        tdiff = tnow - tstart;

        pthread_mutex_lock(&slock);
        xtup   = nth_up;
        xtexp  = nth_expect;
        xserr  = nss_err;
        xsfine = nss_fine;
        xsexp  = iter_expect*nth_expect;
        pthread_mutex_unlock(&slock);

        gsprog = ((double)(xsfine+xserr)*100.0)/xsexp;

        pthread_mutex_lock(&info_lock);
        gsum = (long long) download_sum;
        glen = download_elen;
        pthread_mutex_unlock(&info_lock);

        gesum = glen * xtexp * iter_expect;
        gbprog = ((double)gsum*100.0)/gesum;

#if 0
        Trace("len %lld gesum %lld gbprog %f\n",
            (long long)glen, (long long)gesum, gbprog);
#endif

#if 0
        Trace("elap %5d/%d up %d sum %lld, len %lld\n",
            tdiff, pg_timeout, xtup, 
            (long long)gsum, glen);
#endif

#if 0
        Trace("elap %5d/%d up %d/%d sum %lld %lld, prog %6.2f %%\n",
            tdiff, pg_timeout, xtup, xtexp,
            (long long)gsum, (long long)gesum, gbprog);
#endif

#if 0
        Trace(
            "elap %5d/%d th %3d/%d sess %5d/%d/%d byte %12lld/%lld %7.2f %%\n",
            tdiff, pg_timeout,
            xtup, xtexp, xsfine, xserr, xsexp,
            (long long)gsum, (long long)gesum, gbprog);
#endif

#if 0
        Trace(
            "time %5d/%d th %3d/%d sess %5d/%d/%d %7.2f%% byte %12lld/%lld %7.2f%%\n",
            tdiff, pg_timeout,
            xtup, xtexp, xsfine, xserr, xsexp, gsprog,
            (long long)gsum, (long long)gesum, gbprog);
#endif

        if(rcount%5==0) {
            Trace(
            "=     time     th      sess                       byte\n");

        }
        Trace(
            "= %5d/%d %3d/%d %5d/%d/%d %7.2f%% %12lld/%lld %7.2f%%\n",
            tdiff, pg_timeout,
            xtup, xtexp, xsfine, xserr, xsexp, gsprog,
            (long long)gsum, (long long)gesum, gbprog);

        rcount++;


        if(tdiff>pg_timeout) {
            Error("Timeout; we tier to wait");
            break;
        }
#if 0
        if(xtup<xtexp) {
            Error("sorry, too small thread (expected %d, upped %d)",
                xtexp, xtup);
            exit(43);
        }
#endif

        if(xtup==0) {
            Trace("complete\n");
            if(xsfine==xsexp) {
                Trace("It seems perfect\n");
            }   
            break;
        }   

        sleep(mon_interval);
    }

    tend = time(NULL);
    tdiff = tend - tstart;
    Trace("elap %6ds ; limit %ds\n", tdiff, pg_timeout);

    exit(0);
}





