/*
 * Copyright (C) 2002-2005 StarBED Project All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/*
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/types.h>
#include <fcntl.h>
#include <string.h>
#ifdef __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif
#include <time.h>
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifdef HAVE_SYSLOG_H
#include <syslog.h>
#endif
#include <errno.h>

#ifndef LOG_LINESIZE
#define LOG_LINESIZE    BUFSIZ
#endif


volatile int use_syslog = 1;

pthread_mutex_t trace_lock = PTHREAD_MUTEX_INITIALIZER;

#define XPID    getpid()
#define XTID    pthread_self()


/*
 * RawTrace - show message and flush it.
 *          main body of Trace() (see msg.h).
 */

#ifdef __STDC__
int
RawTrace(const char *fmt, ...)
#else
int
RawTrace(va_alist)
va_dcl
#endif
{
#ifndef __STDC__
    char *fmt;
#endif
    va_list args;

#ifdef __STDC__
    va_start(args,fmt);
#else
    va_start(args);
    fmt = va_arg(args, char *);
#endif

    MUTEX_LOCK(&trace_lock);

    vfprintf(stderr,fmt,args);
    fflush(stderr);

    MUTEX_UNLOCK(&trace_lock);

    va_end(args);

    return 0;
}

/*
 * Log - save message with time stamp as log and shows it
 */

#ifdef __STDC__
int
Log(const char *fmt, ...)
#else
int
Log(va_alist)
va_dcl
#endif
{
#ifndef __STDC__
    char *fmt;
#endif
    va_list args;
    char rmsg[LOG_LINESIZE];
    char msg[LOG_LINESIZE];
    struct tm tmbuf;
    time_t cur;

#ifdef __STDC__
    va_start(args,fmt);
#else
    va_start(args);
    fmt = va_arg(args, char *);
#endif

    cur = time(NULL);
    localtime_r(&cur, &tmbuf);

    vsprintf(rmsg, fmt, args);
    strftime(msg, LOG_LINESIZE, "%Y-%m-%d_%H:%M:%S ", &tmbuf);
    if(strlen(msg)+strlen(rmsg)+1<LOG_LINESIZE) {
        strcat(msg, rmsg);
    }
    else {
        strcpy(msg, rmsg);
    }

    syslog(LOG_NOTICE, msg);

    MUTEX_LOCK(&trace_lock);

    fprintf(stderr, "LOG %s\n", msg);
    fflush(stderr);

    MUTEX_UNLOCK(&trace_lock);


    va_end(args);

    return 0;
}


/*
 * Error - show messages with string-"ERROR" and process/thread-IDs
 */

#ifdef __STDC__
int
Error(const char *fmt, ...)
#else
int
Error(va_alist)
va_dcl
#endif
{
#ifndef __STDC__
    char *fmt;
#endif
    va_list args;
    char rmsg[LOG_LINESIZE];
    char msg[LOG_LINESIZE];
    struct tm tmbuf;
    time_t cur;

#ifdef __STDC__
    va_start(args,fmt);
#else
    va_start(args);
    fmt = va_arg(args, char *);
#endif

    cur = time(NULL);
    localtime_r(&cur, &tmbuf);

    vsprintf(rmsg, fmt, args);
    sprintf(msg, "### ERROR %d:%08x> ", XPID, XTID);
    if(strlen(msg)+strlen(rmsg)+1<LOG_LINESIZE) {
        strcat(msg, rmsg);
    }
    else {
        strcpy(msg, rmsg);
    }

    MUTEX_LOCK(&trace_lock);

    fprintf(stderr, "%s\n", msg);
    fflush(stderr);

    MUTEX_UNLOCK(&trace_lock);

    syslog(LOG_ERR, msg);

    va_end(args);

    return 0;
}

#ifdef __STDC__
int
Warn(const char *fmt, ...)
#else
int
Warn(va_alist)
va_dcl
#endif
{
#ifndef __STDC__
    char *fmt;
#endif
    va_list args;
    char rmsg[LOG_LINESIZE];
    char msg[LOG_LINESIZE];
    struct tm tmbuf;
    time_t cur;

#ifdef __STDC__
    va_start(args,fmt);
#else
    va_start(args);
    fmt = va_arg(args, char *);
#endif

    cur = time(NULL);
    localtime_r(&cur, &tmbuf);

    vsprintf(rmsg, fmt, args);
    sprintf(msg, "--- WARN %d:%08x> ", XPID, XTID);
    if(strlen(msg)+strlen(rmsg)+1<LOG_LINESIZE) {
        strcat(msg, rmsg);
    }
    else {
        strcpy(msg, rmsg);
    }

    MUTEX_LOCK(&trace_lock);

    fprintf(stderr, "%s\n", msg);
    fflush(stderr);

    MUTEX_UNLOCK(&trace_lock);

    syslog(LOG_WARNING, msg);

    va_end(args);

    return 0;
}



