/*
 * Copyright (C) 2002-2005 StarBED Project All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/*
 *  nio.c - network I/O procedures
 *  $Id: nio.c,v 1.1 2006/11/27 01:56:36 k-chinen Exp $
 */


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#ifdef __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#ifdef HAVE_POLL_H
#include <poll.h>
#endif
#ifdef HAVE_SYS_POLL_H
#include <sys/poll.h>
#endif
#include <time.h>
#include <sys/wait.h>
#include <sys/resource.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <errno.h>

#include "msg.h"
#include "hname.h"
#include "sman.h"
#include "socket.h"

/*
#define show_addr(m,a)      raw_show_addr(m,a)
*/

volatile int nio_trace=0;
#define Trace               if(nio_trace) RawTrace

int msleep(int x);

#ifndef MIN
#define MIN(x,y)    ((x)<(y) ? (x) : (y))
#endif



#ifndef LOG_LINESIZE
#define LOG_LINESIZE    (BUFSIZ)
#endif

int
fdprintf(const int fd, const char *fmt, ...)
{
    va_list args;
    char buffer[LOG_LINESIZE];
    int nwrite;
    int enwrite;

    va_start(args,fmt);

    if(fd<0) {
        Error("fdprintf: ignore fd# %d", fd);
        goto leave;
    }

    memset(buffer, 0, sizeof(buffer));
    vsprintf(buffer, fmt, args);
    enwrite = strlen(buffer);
try_write:
    nwrite = write(fd, buffer, enwrite);
#if 0
    Trace("fdprintf: fd# %d, wroten %d, expected %d\n",
        fd, nwrite, enwrite);
#endif
    if(nwrite!=enwrite) {
        Warn("fdprintf: fd# %d, wroten %d, expected %d",
            fd, nwrite, enwrite);
    }
    if(nwrite<0) {
        if(errno==EINTR) {
            goto try_write;
        }
        else {
            Error("fdprintf: fail write() fd# %d (%d)", fd, errno);
        }
    }

leave:
    va_end(args);

    return nwrite;
}

/*
 * NOTE:
 *  this routine optimize to RM.
 *  you have to care if you use other target.
 */

/*
 *  success 0
 *  fail    <0
 *  timeout >0
 *
 */
int
nio_rmqa(int rfd, char *src, int *rp, char *dst, int dlen)
{
    int ret;
    int nr;
    int sr;
    int ck;
    int cc;
    int nw;
    int trp;
    int tt;

    char line[BUFSIZ];
    char *lp;
    register char *p, *q;

    trp = -1;
    
    memset(dst, 0, dlen);
    *rp = -1;

    if(nio_trace) {
        dump_string("nio_rmqa Q", src);
    }
    nw = write(rfd, src, strlen(src));
    if(nw<0) {
        Error("nw neg %d; fd# %d", nw, rfd);
        goto fail;
    }
    else {
        Trace("nw %d; fd# %d\n", nw, rfd);
    }

    nr = 0;
    sr = 0;
    cc = 0;
    while(1) {
#if 0
        ck = isreadable_via_select(rfd, 0, 500000);
#endif
        ck = waitread(rfd, MS(500));
        if(ck) {
try_read:
            nr = read(rfd, &dst[sr], dlen-sr-1);
            if(nr==0) {
                Trace("reach to EOF\n");
                break;
            }
            if(nr<0) {
                if(errno==EINTR)
                    goto try_read;
                Error("nio_rmqa: fail read() (%d)", errno);
                goto fail;
                break;
            }
#if 0
            Trace("nr %d\n", nr);
#endif
            sr += nr;
            dst[sr] = '\0';
#if 0
            dump_string("dst-mid", dst);
#endif
            cc = 0;
        } 
        else {
            Trace("connection is cold\n");
            cc++;
        }

        if(sr>3 && trp<0) {
            tt = atoi(dst);
            if(tt>0) {
                trp = tt;
            }
            else {
                Error("strange response code %d '%s'\n", tt, dst);
                goto fail;
            }
        }

        if(trp>0) {
            /* multi line */
            if(trp%100==1) {

                lp = dst;
                while(1) {
                    lp = sgets(lp, line, BUFSIZ);
                    if(line[0]=='\0') {
                        break;
                    }
                    if(line[0]=='.') {
                        Trace("reach to EOM; nr %d, sr %d\n", nr, sr);
                        goto success;
                    }
                } 

            }
            /* single line */
            else {
                q = NULL;
                p = dst;
                while(*p) {
                    q = p;
                    p++;
                }

                /* got LF --- line was complete */
                if(*q=='\n') {
                    goto success;
                }
            } 

        }
        /* response code is not clear -> too short or empty input */
        else {
        }
        
#if 0
        msleep(100);
#endif
    }

    /* maybe timeout */
    if(sr==0) {
        Error("no data read <cc %d, sr %d>", cc, sr);
        goto timeout;
    }

    goto success;

fail:
    ret = -1;
    goto out;

timeout:
    ret = 1;
    goto out;

success:
    ret = 0;
    *rp = atoi(dst);
    goto out;

out:
#if 0
    dump_string("dst-fin", dst);
#endif
    if(nio_trace) {
        Trace("rcode %d ; length %d/%d; ret %d\n",
            *rp, strlen(dst), dlen, ret);
        dump_stringchunk("nio_rmqa A", dst);
    }

    return ret;
}




/*
 * fdbgets - fd buffering get string  
 *
 * return 
 *      -1  error
 *       0  end
 *      >0  success, the length of line
 */
/* NOTE: this routine supports until BUFSIZ */

int
fdbgets(int fd, char *bbuf, char *line)
{
    int ret;
    int nlen;
    int c;
    int ck;
    int nr;
    int w;
    register char *p;
    register char *z;

start:
    if(nio_trace>5) {
        dump_stringchunk("bbuf", bbuf);
    }

    p = bbuf;
    z = NULL;
    c = 0;
    while(c<BUFSIZ-1&&*p) {
        if(!z && *p=='\n') {
            z = p;
        }
        p++;
        c++;
    }
    if(c==BUFSIZ) {
        return -1;
    }
    
    nlen = c;
    if(nio_trace>6) {
        fprintf(stderr, " nlen %d ; z %p %d\n", nlen, z, z-bbuf);
    }

    if(!z) {
        if(nlen>=BUFSIZ-1) {
            if(nio_trace>6) {
                fprintf(stderr, "  buffer is full\n");
            }
            goto full;
        }

        if(nio_trace>6) {
            fprintf(stderr, "  want new input\n");
            fprintf(stderr, "  isreadable? fd# %d,\n", fd);
        }
        ck = isreadable_via_select(fd, 1, 0);
        if(nio_trace>6) {
            fprintf(stderr, "  ck %d\n", ck);
        }
        if(ck) {
            if(nio_trace>6) {
                fprintf(stderr, "  read in %d+%d\n",
                    nlen, BUFSIZ-nlen-1);
            }
try_read:
            nr = read(fd, &bbuf[nlen], BUFSIZ-nlen-1);
            if(nio_trace>6) {
                fprintf(stderr, "  nr %d\n", nr);
            }
            if(nr<0) {
                if(errno==EINTR)
                    goto try_read;
                Error("fdbgets: fail read (%d)", errno);
                return -1;
            }
            if(nr==0) {
                goto eos;
            }

            nlen += nr;
            bbuf[nlen] = '\0';

            if(nio_trace>6) {
                dump_stringchunk("tmp ", bbuf);
            }
        }

        if(nio_trace>6) {
            fprintf(stderr, "restart\n");
        }
        goto start;
    }

    w = z - bbuf + 1;
    if(nio_trace>6) {
        fprintf(stderr, " w %d\n", w);
    }

    memset(line, 0xff, BUFSIZ);
    memcpy(line, &bbuf[0], w);
    line[w] = '\0';

    memmove(&bbuf[0], &bbuf[w], BUFSIZ-w);

    if(nio_trace>5) {
        dump_stringchunk("head", line);
        dump_stringchunk("rest", bbuf);
    }

    ret = w;

    goto out;

full:
    memset(line, 0xff, BUFSIZ);
    w = MIN(nlen, BUFSIZ-1);
    memcpy(line, &bbuf[0], w);
    line[w] = '\0';
    memset(bbuf, 0, BUFSIZ);    /* clear */
    ret = w;

    goto out;

eos:
    /* end of stream */
    memset(line, 0xff, BUFSIZ);
    w = MIN(nlen, BUFSIZ-1);
    memcpy(line, &bbuf[0], w);
    line[w] = '\0';
    memset(bbuf, 0, BUFSIZ);    /* clear */
    ret = 0;

    goto out;

out:

    return ret;
}

