/*
 * Copyright (C) 2002-2006 StarBED Project All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/*
 * network object retrieval procedures
 * Ken-ichi Chinen
 *
 * $Id: nret.c,v 1.6 2007/06/06 06:02:12 k-chinen Exp $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <time.h>
#include <sys/time.h>
#include <syslog.h>
#include <errno.h>

#ifdef __FreeBSD__
#include <sys/disklabel.h>
#endif

#include <zlib.h>

/*
#include "range.h"
#include "globals.h"
#include "ffp.h"
#include "xmft.h"
*/

#include "msg.h"
#include "socket.h"
#include "http.h"
#include "sman.h"
#include "nio.h"

/* XXX */
#include "fmisc.c"

#ifndef L_PATH
#define L_PATH  BUFSIZ
#endif

pthread_mutex_t info_lock = PTHREAD_MUTEX_INITIALIZER;

int download_report=0;
int download_new=0;
struct timeval download_start;
struct timeval download_end;
long long download_sum=0;
long long download_elen=0;

int upload_report=0;
int upload_new=0;
struct timeval upload_start;
struct timeval upload_end;
long long upload_sum=0;
long long upload_elen=0;

int ftp_wait_welcome_first = 500;
int ftp_wait_welcome_int   = 100;

volatile int nret_trace = 0;
volatile int blkcheck = 1;

#define Trace   if(nret_trace) RawTrace

#ifndef PASS
#define PASS    \
    fprintf(stderr, "PASS %s:%d\n",__FILE__,__LINE__); fflush(stderr)
#endif

#define SUB_MSEC(x,y) \
        ( (x.tv_sec  - y.tv_sec )*1000 + (x.tv_usec - y.tv_usec)/1000 )

#define MIN(x,y)    ((x)<(y) ? (x):(y))



static
int
path_project(char *dst, char *src)
{
    strcpy(dst, src);
}

int
is_gzfile(char *path)
{
    int ret;
    register char *p;

    ret = 0;
    p = path;
    while(*p) {
        if(*p=='.') {
            if(*(p+1)=='Z'&&*(p+2)=='\0') {
                ret = 1;
                goto out;
            }
            if(*(p+1)=='g'&&*(p+2)=='z'&&*(p+3)=='\0') {
                ret = 1;
                goto out;
            }
        }
        p++;
    }
out:
    return ret;
}

int
is_devfile(char *path)
{
    if(strncmp(path,"/dev/", 5)==0) {
        return 1;
    }
    return 0;
}


int
disable_disklabel_protect(int ffd)
{
    Trace("try disable disklabel protect for fd# %d\n", ffd);
#ifdef __FreeBSD__
    {
        volatile int flag;
        int xchk;

        Trace("  FreeBSD DIOCWLABEL\n");
        flag = 1;
        xchk = ioctl(ffd, DIOCWLABEL, &flag);
        if(xchk<0) {
            Error("fail ioctl to disable disklabel protect (%d)", errno);
            return -1;
        }
        else {
            Trace("disklabel protect is disable\n");
        }
    }
#else
    Trace("no way for disklabel protecting\n");
#endif
    return 0;
}

int
enlarge_sockbuf(int dfd, int bufid)
{
    char data_addrstr[BUFSIZ];
    int data_port;
    struct sockaddr_in data_addr;
    int cursize, cslen;
    int firstsize, fslen;
    int trysize;
    int lastsize;
    int i;
    int chk;

#if 0
    Trace("enlarge_sockbuf: fd# %d\n", dfd);
#endif
    if(bufid==SO_RCVBUF) {
        Trace("enlarge_sockbuf: fd# %d, RCVBUF\n", dfd);
    }
    else
    if(bufid==SO_SNDBUF) {
        Trace("enlarge_sockbuf: fd# %d, SNDBUF\n", dfd);
    }
    else {
        Trace("enlarge_sockbuf: fd# %d, bufid# %d\n", dfd, bufid);
    }

    fslen = sizeof(firstsize);
    chk = getsockopt(dfd, SOL_SOCKET, bufid,
            (void*)&firstsize, &fslen);
    if(chk<0) {
        Warn("fail first  getsockopt (%d)", errno);
        return -1;
    }
    else {
        Trace("first buf#%d %d\n", bufid, firstsize);
    }

    Trace("  %9s : %9s %9s %+9s : %9s %+9s\n",
        "try", "cur", "last", "diff", "first", "diff"
    );

    lastsize = -1;
    for(i=1;i<=1024;i++) {
        trysize = i*firstsize;
        chk = setsockopt(dfd, SOL_SOCKET, bufid,
                (const void*)&trysize, sizeof(trysize));
        if(chk<0) {
            Warn("fail setsockopt (%d)", errno);
        }

        cslen = sizeof(cursize);
        chk = getsockopt(dfd, SOL_SOCKET, bufid,
                (void*)&cursize, &cslen);
        if(chk<0) {
            Warn("fail middle getsockopt (%d)", errno);
        }
        Trace("  %9d * %9d %9d %+9d : %9d %+9d\n",
            trysize, cursize, lastsize, lastsize - cursize,
            firstsize, cursize - firstsize);

        if(lastsize>0 && cursize==lastsize) {
            break;
        }

        lastsize = cursize;
    }

    lastsize = ((lastsize+1023)/1024)*1024;
    Trace("  %9d KB-fix\n", lastsize);

    for(i=1;i<=1024;i++) {
        trysize = lastsize+1024;
        chk = setsockopt(dfd, SOL_SOCKET, bufid,
                (const void*)&trysize, sizeof(trysize));
        if(chk<0) {
            Warn("fail setsockopt (%d)", errno);
        }

        cslen = sizeof(cursize);
        chk = getsockopt(dfd, SOL_SOCKET, bufid,
                (void*)&cursize, &cslen);
        if(chk<0) {
            Warn("fail middle getsockopt (%d)", errno);
        }
        Trace("  %9d + %9d %9d %+9d : %9d %+9d\n",
            trysize, cursize, lastsize, lastsize - cursize,
            firstsize, cursize - firstsize);

        if(lastsize>0 && cursize==lastsize) {
            break;
        }

        lastsize = cursize;
    }

    cslen = sizeof(cursize);
    chk = getsockopt(dfd, SOL_SOCKET, bufid,
            (void*)&cursize, &cslen);
    Trace("last  buf#%d %d\n", bufid, cursize);
    if(chk<0) {
        Warn("fail last   getsockopt (%d)", errno);
    }

    return 0;
}


int
remove_dirpart(char *dst, char *src)
{
    register char *p, *q;
    register int c;
    char *r;

#if 1
    Trace("remove_dirpath: src '%s'\n", src);
#endif

    r = NULL;
    p = src;
    while(*p) {
        if(*p=='/') {
            r = p;
        }
        p++;
    }
    if(r) {
#if 0
        Trace("last path '%s'\n", r);
#endif
        p = r;
        p++;    /* skip '/' */
    }
    else {
        Trace("remove_dirpart: no path\n");
        p = src;
    }
    q = dst;
    c = 0;
    while(*p && c<BUFSIZ) {
        *q++ = *p++;
        c++;
    }
    *q = '\0';

#if 1
    Trace("remove_dirpath: dst '%s'\n", dst);
#endif

    return 0;
}

int
FTP_put(int cfd, char *user, char *pass, char *rpath, char *lpath)
{
    char arpath[BUFSIZ];
    char rfile[BUFSIZ];
    char lfile[BUFSIZ];
    char buf[BUFSIZ];
    int blen;
    int nw, nr;
    int chk;
    int dfd, ffd;
    char ct_str[L_PATH], date_str[L_PATH], lmd_str[L_PATH];
    char pub_str[L_PATH];
    struct stat sbuf;
    off_t len;
    int ret;
    int usez=0;
    int ba=0;
    gzFile gzf;

    Trace("FTP_put: dst rpath '%s', user '%s', pass '%s'\n", 
        rpath, user, pass);
    Trace("         src lpath '%s'\n", lpath);

    ret = 0;
    len = (off_t) 0;

    ffd = open(lpath, O_RDONLY);
    if(ffd < 0) {
        Error("fail open() local '%s' (%d)", lpath, errno);
        goto nofile;
    }

    chk = stat(lpath, &sbuf);
    if(chk) {
        Error("fail stat() local '%s' (%d)", lpath, errno);
        goto nofile;
    }
    if(S_ISDIR(sbuf.st_mode)) {
        Error("local is directory '%s' (%d)", lpath, errno);
        goto isdir;
    }

#if 1
    MUTEX_LOCK(&info_lock);
    upload_elen = sbuf.st_size;
    MUTEX_UNLOCK(&info_lock);
#endif

    remove_dirpart(lfile, lpath);
    remove_dirpart(rfile, rpath);

    Trace("  place   %-40s %s\n", "path", "file");
    Trace("  local   %-40s %s\n", lpath, lfile);
    Trace("  remote  %-40s %s\n", rpath, rfile);
    if(!rfile[0]) {
        Trace(
"                                               *** copy from local ***\n");
        strcpy(rfile, lfile);
        Trace("  remote* %-40s %s\n", rpath, rfile);
    }
    strcpy(arpath, rfile);

    if(is_devfile(lpath)) {
        ba = 1;
    }
    if(is_gzfile(rpath)) {
        usez = 1;
    }


try_first_read:
    nr = read(cfd, buf, BUFSIZ);
    if(nr < 0) {
        if(errno == EINTR)
            goto try_first_read;
    }
    if(strncasecmp(buf, "220", 3) != 0) {
        goto skip_data_conn;
    }

    fdprintf(cfd, "USER %s\r\n", user);
try_user_read:
    nr = read(cfd, buf, BUFSIZ);
    if(nr < 0) {
        if(errno == EINTR)
            goto try_user_read;
    }
    if(strncasecmp(buf, "331", 3) != 0) {
        goto skip_data_conn;
    }

    fdprintf(cfd, "PASS %s\r\n", pass);
try_pass_read:
    nr = read(cfd, buf, BUFSIZ);
    if(nr < 0) {
        if(errno == EINTR)
            goto try_pass_read;
    }
    if(strncasecmp(buf, "230", 3) != 0) {
        goto skip_data_conn;
    }


    /*
     * process directories in remote path
     */
 {
    char dname[BUFSIZ];
    register char *p, *q, *r;
    register int c;

    r = rpath;
    while(*r) {
        p = r;
        c = 0;
        while(*p&&*p!='/') {
            p++;
            c++;
        }

        /* have slash - directory? */
        if(*p=='/') {
            p = r;
            q = dname;
            c = 0;
            while(*p&&*p!='/'&&c<BUFSIZ) {
                *q++ = *p++;
                c++;
            }
            *q = '\0';

            Trace("  find dir '%s'\n", dname);

            fdprintf(cfd, "CWD %s\r\n", dname);
try_chdir_read:
            nr = read(cfd, buf, BUFSIZ);
            if(nr < 0) {
                if(errno == EINTR)
                    goto try_chdir_read;
            }
            if(strncasecmp(buf, "250", 3) == 0) {
                Trace("    success chdir\n");
                goto direxist;
            }
            Trace("    fail chdir\n");

            fdprintf(cfd, "MKD %s\r\n", dname);
try_mkdir_read:
            nr = read(cfd, buf, BUFSIZ);
            if(nr < 0) {
                if(errno == EINTR)
                    goto try_mkdir_read;
            }
            Trace("got :%s", buf);
            if(strncasecmp(buf, "257", 3) == 0) {
                Trace("    success mkdir\n");
            }
            else {
                Error("server reject MKD <%s>", buf);
                goto skip_data_conn;
            }

            fdprintf(cfd, "CWD %s\r\n", dname);
try_chdir2_read:
            nr = read(cfd, buf, BUFSIZ);
            if(nr < 0) {
                if(errno == EINTR)
                    goto try_chdir2_read;
            }
            if(strncasecmp(buf, "250", 3) == 0) {
                Trace("    success chdir again\n");
            }
            else {
                Error("server reject second CWD <%s>", buf);
                goto skip_data_conn;
            }

direxist:
            (void)0;

            p++;
            r = p;
        }
        /* no slash */
        else {
            Trace("  reach end\n");
            break;
        }
    }

  }

    fdprintf(cfd, "TYPE I\r\n");
try_type_read:
    nr = read(cfd, buf, BUFSIZ);
    if(nr < 0) {
        if(errno == EINTR)
            goto try_type_read;
    }
    if(strncasecmp(buf, "200", 3) != 0) {
        goto skip_data_conn;
    }


    fdprintf(cfd, "PASV\r\n");
try_pasv_read:
    nr = read(cfd, buf, BUFSIZ);
    if(nr < 0) {
        if(errno == EINTR)
            goto try_pasv_read;
    }
    if(strncasecmp(buf, "227", 3) != 0) {
        goto skip_data_conn;
    }

    {
        char data_addrstr[BUFSIZ];
        int data_port;
        struct sockaddr_in data_addr;

        chk = parse_ftp_peer(buf, data_addrstr, &data_port);

        Trace("data conn addr '%s' port %d\n", data_addrstr, data_port);

        chk = addr_bynumber(data_addrstr, data_port, &data_addr);

        dfd = TCP_connect_addr(&data_addr, data_port);
    }


#if 0
    dfd = -1;
    goto skip_data_conn;
#endif


    if(dfd < 0) {
        goto skip_data_conn;
    }
    Trace("data connection fd# %d\n", dfd);


    Trace("; usez %d, ba %d, arpath '%s'\n", usez, ba, arpath);
    fdprintf(cfd, "STOR %s\r\n", arpath);
try_store_read:
    nr = read(cfd, buf, BUFSIZ);
    if(nr < 0) {
        if(errno == EINTR)
            goto try_store_read;
    }
    if(strncasecmp(buf, "150", 3) != 0) {
        Error("server reject STOR <%s>", buf);
        goto skip_data_conn;
    }
    Trace("  success STOR\n");

#if 1
    MUTEX_LOCK(&info_lock);
    upload_report = 1;
    gettimeofday(&upload_start, NULL);
    upload_end.tv_sec = 0;
    upload_end.tv_usec = 0;
    MUTEX_UNLOCK(&info_lock);
#endif

    gzf = gzdopen(dfd, "wb");
    if(gzf==NULL) {
        Error("fail gzdopen for FTP STOR");
    }


    while(1) {
try_read:
        nr = read(ffd, buf, BUFSIZ);
        if(nr < 0) {
            if(errno == EINTR) {
                goto try_read;
            }
            else {
                Error("fail read() fd# %d, (%d)", ffd, errno);
                close(ffd);
                if(usez) {
                    gzclose(gzf);
                }
                close(dfd);
                goto skip_data_conn;
            }
        }
        if(nr == 0) {
            break;
        }

        if(usez) {
            nw = gzwrite(gzf, buf, nr);
            if(nw < 0) {
                Error("fail gzwrite() for FTP STOR");
                close(ffd);
                gzclose(gzf);
                close(dfd);
                goto skip_data_conn;
            }
            if(nw==0) {
                break;
            }
            
        }
        else {

try_write:
            nw = write(dfd, buf, nr);
            if(nw < 0) {
                if(errno == EINTR) {
                    goto try_write;
                }
                Error("FTP-put fail write() fd# %d, (%d)", cfd, errno);
                close(ffd);
                close(dfd);
                goto skip_data_conn;
            }

        }

        len += nw;
#if 0
        Trace(" %4d/%4lld\n", nw, (long long)len);
#endif
        MUTEX_LOCK(&info_lock);
        upload_sum = (long long)len;
        MUTEX_UNLOCK(&info_lock);

#if 0
        usleep(30000);
#endif

    }

end:
    close(ffd);
    if(usez) {
        gzclose(gzf);
    }
    close(dfd);
    shutdown(dfd, 2);

skip_data_conn:
    fdprintf(cfd, "quit\r\n");
    close(cfd);
    shutdown(cfd, 2);

    MUTEX_LOCK(&info_lock);
    gettimeofday(&upload_end, NULL);
    MUTEX_UNLOCK(&info_lock);

    goto out;

isdir:
nofile:
    Error("it is directory or not found");

    ret = 1;

out:
    Trace("lpath '%s' len %lld\n", lpath, (long long) len);
    return ret;
}


#if 0
static
char*
skipwhite(char *p)
{
    while(*p==' ' ||*p=='\t') {
        p++;
    }
    return p;
}
#endif

#define SKIPWHITE   {while(*p==' '||*p=='\t') {p++;}}
#define STORESTR \
    {c=0;\
    while(c<BUFSIZ&&*p&&*p!=' '&&*p!='\t'&&*p!='\r'&&*p!='\n') \
    {*q++=*p++;c++;} *q='\0';}


long long
parse_filesize_inlistofftp(char *line)
{
    long long ret;
    register int c;
    register char *p, *q;
    char token[BUFSIZ];
    int i;

    ret = -1;

    i = 0;
    p = line;
    while(*p) {
        SKIPWHITE;
        q = token;
        STORESTR;
#if 0
        fprintf(stderr, " token %2d: '%s'\n", i, token);
#endif
        if(i==4) {
#ifdef HAVE_ATOLL
            ret = atoll(token);
#else
            ret = (long long)atol(token);
#endif
        }

        if(token[0]=='\0') {
            break;
        }
        i++;
    }

    return ret;
}

static
int
iocheck(int ffd)
{
    char xbuf[BUFSIZ];
    off_t xr, xw;
    off_t xk;

    Trace("X check disk write/reading\n");
    xk = lseek(ffd, 0, 0);
    if(xk<0) {
        Error("X fail lseek 1 (%d)", errno);
        goto cannot_write;
    }

    xr = read(ffd, xbuf, BUFSIZ);
    Trace("X xr %lld\n", (long long)xr);
    if(xr<0) {
        Error("X fail read (%d)", errno);
        goto cannot_write;
    }

    xk = lseek(ffd, 0, 0);
    if(xk<0) {
        Error("X fail lseek 2 (%d)", errno);
        goto cannot_write;
    }
    
    xw = write(ffd, xbuf, xr);
    Trace("X xw %lld\n", (long long)xw);
    if(xw<0) {
        Error("X fail write (%d)", errno);
        goto cannot_write;
    }

    xk = lseek(ffd, 0, 0);
    if(xk<0) {
        Error("X fail lseek 3 (%d)", errno);
        goto cannot_write;
    }

    return 0;

cannot_write:
    return -1;
}

/*
 * returns fd for specified path's file
 *
 */
int
assign_path(char *lpath, int *isdev)
{
    int ret;
    int ffd;
    struct stat sbuf;
    int chk;

    ret = -1;

    ffd = open(lpath, O_WRONLY);
    if(ffd<0) {
        Error("fail open() 1 '%s' (%d)", lpath, errno);
        ffd = open(lpath, O_WRONLY|O_CREAT|O_TRUNC, 0644);
        if(ffd < 0) {
            Error("fail open() 2 '%s' (%d)", lpath, errno);
            goto nofile;
        }
    }
    chk = stat(lpath, &sbuf);
    if(chk) {
        Error("fail stat() '%s' (%d)", lpath, errno);
        goto nofile;
    }
    if(S_ISDIR(sbuf.st_mode)) {
        Error("it is directory '%s' (%d)", lpath);
        goto isdir;
    }

    ret = 0;

    if(is_devfile(lpath)) {
        Trace("it is device file\n");
        *isdev = 1;
    }
    else {
        *isdev = 0;
    }
    Trace("; isdev %d\n", *isdev);

    goto out;

nofile:
isdir:

out:
    return ret;
}

#ifndef XBUFSIZ
#define XBUFSIZ (32*1024)
#endif
#ifndef RCREPINT
#define RCREPINT (100)
#endif

/*
int g_blen = XBUFSIZ/2;
*/
int g_blen = XBUFSIZ;
int rbsize = XBUFSIZ;

int
FTP_get(int cfd, char *user, char *pass, char *rpath, char *lpath,
    off_t *hsaved)
{
    char buf[BUFSIZ];
    char xbuf[XBUFSIZ];
    int blen;
    int pnr;
    int nw, nr;
    int chk;
    int dfd, ffd;
    char ct_str[L_PATH], date_str[L_PATH], lmd_str[L_PATH];
    char pub_str[L_PATH];
    struct stat sbuf;
    off_t len;
    int ret;
    long long elen;
    int usez=0;
    int ba=0;
    gzFile gzf;
    int cc;
    int mrt, mwt;
    struct timeval mrst, mret;
    struct timeval mwst, mwet;
    int rcn, rci;
    int wcn, wci;
    int rpcn, rpci;
    int isdev=0;

    Trace("FTP_get: user '%s', pass '%s'\n\
  rpath '%s'\n\
  lpath '%s'\n",
        user, pass, rpath, lpath);

    ret = 0;
    len = (off_t) 0;
    elen = -1;

#if 0
    Trace(" XBUFSIZ %d, g_blen %d\n", XBUFSIZ, g_blen);
#endif
    blen = MIN(XBUFSIZ, g_blen);
#if 0
    Trace(" blen %d\n", blen);
#endif
    blen = (blen/512)*512;
    if(blen<512)
        blen = 512;
    Trace(" normalized blen %d\n", blen);

#if 1
    ffd = open(lpath, O_WRONLY);
    if(ffd<0) {
        Warn("fail open() local for append 1 '%s' (%d)", lpath, errno);
        ffd = open(lpath, O_WRONLY|O_CREAT|O_TRUNC, 0644);
        if(ffd < 0) {
            Error("fail open() local for create 2 '%s' (%d)", lpath, errno);
            goto nofile;
        }
    }

    chk = stat(lpath, &sbuf);
    if(chk) {
        Error("fail stat() local '%s' (%d)", lpath, errno);
        goto nofile;
    }
    if(S_ISDIR(sbuf.st_mode)) {
        Error("local is directory '%s' (%d)", lpath);
        goto isdir;
    }

    ba = 0;
    if(is_devfile(lpath)) {
        Trace("local '%s' is device file (guessed)\n", lpath);
        if(blkcheck) {
            Trace("enter block align\n");
            ba = 1;
        }
    }   
#else
    ffd = assign_path(lpath, &isdev);
#endif

    if(ffd<0) {
        Error("fail open local file");
    }

    if(isdev) {
        Trace("local '%s' is device file (specified)\n", lpath);
        if(blkcheck) {
            Trace("enter block align\n");
            ba = 1;
        }
    }

#if 0
    usez = 0;
    if(is_gzfile(rpath)) {
        Trace("we enter gz mode\n");
        usez = 1;
    }
#endif

    /* when both pathes are gzfile, it means direct copy */
    usez = 0;
    if(is_gzfile(rpath)) {
        if(is_gzfile(lpath)) {
            Trace("gz to gz, no gz mode\n");
        }
        else {
            Trace("we enter gz mode\n");
            usez = 1;
        }
    }


    Trace("; usez %d, ba %d\n", usez, ba);

    /*
     *  thru welcome messages
     */
#if 0
    /*
     *  expect short message
     */

try_first_read:
    nr = read(cfd, buf, BUFSIZ);
    if(nr < 0) {
        if(errno == EINTR)
            goto try_first_read;
    }
#if 0
    dump_stringsegment("S:", buf, nr);
#endif
    if(strncasecmp(buf, "220-", 4) == 0) {
        goto try_first_read;
    }

    if(strncasecmp(buf, "220", 3) != 0) {
        goto skip_data_conn;
    }

#endif

    /*
     *  expect long message
     */
  {
    int   sr;
    int   re;
    char *p;
    char  line[BUFSIZ];
#if 0
    int   ftp_wait_welcome_first = 500;
    int   ftp_wait_welcome_int   = 100;
#endif

    chk = waitread_via_poll(cfd, MS(ftp_wait_welcome_first));
    if(chk==0) {
        Trace("skip first message\n");
        goto skip_first_message;
    }
    if(chk<0) {
        Error("strange wait_read");
    }
    Trace("have first message; read them\n");

    sr = 0;
    while ((chk = waitread_via_poll(cfd, MS(ftp_wait_welcome_int)))>0)  {
        re = BUFSIZ - sr - 1;
        Trace("read message %d %d\n", sr, re);

try_first_read:
        nr = read(cfd, &buf[sr], re);
#if 1
        Trace("nr %d\n", nr);
#endif
        if(nr < 0) {
            if(errno == EINTR)
                goto try_first_read;
        }
        if(nr==0) {
            break;
        }
        sr += nr;
#if 0
        dump_stringsegment("S:", buf, sr);
#endif
    }
    buf[sr] = '\0';
    if(nret_trace) {
        dump_stringchunk("S:", buf);
    }

    p = buf;
    while(1) {
        p = sgets(p, line, L_PATH);
        if(!line[0]) {
            break;
        }
        if(strncasecmp(line, "220-", 4) == 0) {
            continue;
        }
        if(strncasecmp(line, "220 ", 4) == 0) {
            break;
        }
        else {
            Warn("strange response code, it is not 200 and 200- ");
#if 1
            dump_string("line", line);
#endif
            break;
        }
    }

skip_first_message:
    (void)1;


  }

    /*
     * USER
     */
    fdprintf(cfd, "USER %s\r\n", user);
try_user_read:
    nr = read(cfd, buf, BUFSIZ);
    if(nr < 0) {
        if(errno == EINTR)
            goto try_user_read;
        Error("fail read() for USER (%d)", errno);
    }
    if(strncasecmp(buf, "331", 3) != 0) {
        Error("fail FTP USER <%d>", atoi(buf));
        goto skip_data_conn;
    }

    /*
     * PASS
     */
    fdprintf(cfd, "PASS %s\r\n", pass);
try_pass_read:
    nr = read(cfd, buf, BUFSIZ);
    if(nr < 0) {
        if(errno == EINTR)
            goto try_pass_read;
        Error("fail read() for PASS (%d)", errno);
    }
    if(strncasecmp(buf, "230", 3) != 0) {
        Error("fail FTP PASS <%d>", atoi(buf));
        goto skip_data_conn;
    }

#if 0
    Trace("USER, PASS were done\n");
#endif

    /*
     * TYPE I
     */
    fdprintf(cfd, "TYPE I\r\n");
try_type_read:
    nr = read(cfd, buf, BUFSIZ);
    if(nr < 0) {
        if(errno == EINTR)
            goto try_type_read;
        Error("fail read() for TYPE (%d)", errno);
    }
    if(strncasecmp(buf, "200", 3) != 0) {
        Error("fail FTP TYPE <%d>", atoi(buf));
        goto skip_data_conn;
    }
    
    /*
     * SIZE
     */
#if 1
    fdprintf(cfd, "SIZE %s\r\n", rpath);
try_size_read:
    nr = read(cfd, buf, BUFSIZ);
    if(nr < 0) {
        if(errno == EINTR) {
            goto try_size_read;
        }
    }
    dump_stringsegment("size?", buf, nr);
    if(strncasecmp(buf, "213", 3) == 0) {
        elen = (long long)atoll(&buf[4]);

        if(elen>0) {
            Trace("FTP server report its size as %lld\n", elen);
            Trace(" %lldBUF +%lldBYTES\n", elen/blen, elen%blen);
#if 1
            MUTEX_LOCK(&info_lock);
            download_elen = elen;
            MUTEX_UNLOCK(&info_lock);
#endif
        }
    }
    else {
        Error("fail FTP SIZE <%d>", atoi(buf));
        goto nofile;
    }
#endif

    /***
     *** LIST
     ***/
#if 0
    fdprintf(cfd, "PASV\r\n");
try_list_pasv_read:
    nr = read(cfd, buf, BUFSIZ);
    if(nr < 0) {
        if(errno == EINTR) {
            goto try_list_pasv_read;
        }
    }
    if(strncasecmp(buf, "227", 3) != 0) {
        Error("fail FTP PASV for LIST <%d>", atoi(buf));
        goto skip_data_conn;
    }

    {
        char data_addrstr[BUFSIZ];
        int data_port;
        struct sockaddr_in data_addr;

        chk = parse_ftp_peer(buf, data_addrstr, &data_port);

        Trace("data conn addr '%s' port %d\n", data_addrstr, data_port);

        chk = addr_bynumber(data_addrstr, data_port, &data_addr);

        dfd = TCP_connect_addr(&data_addr, data_port);
    }

    if(dfd < 0) {
        Error("cannot get data connection for FTP LIST");
        goto skip_data_conn;
    }
    Trace("data connection fd# %d\n", dfd);

    fdprintf(cfd, "LIST %s\r\n", rpath);
try_list_read:
    nr = read(cfd, buf, BUFSIZ);
    if(nr < 0) {
        if(errno == EINTR) {
            goto try_list_read;
        }
    }
    if(strncasecmp(buf, "150", 3) != 0) {
        Error("fail FTP LIST <%d>", atoi(buf));
        goto skip_check_filesize;
    }

    while(1) {
        nr = read(dfd, buf, BUFSIZ);
#if 0
        Trace("list read %d\n", nr);
#endif
        if(nr<=0) {
            break;
        }
        if(nr>0) {
#if 0
            dump_stringsegment("buf", buf, nr);
#endif
        }
    }
    shutdown(dfd, 2);
    close(dfd);

    elen = parse_filesize_inlistofftp(buf);

    if(elen>0) {
        Trace("FTP server report its size as %lld\n", elen);
#if 1
    MUTEX_LOCK(&info_lock);
    download_elen = elen;
    MUTEX_UNLOCK(&info_lock);
#endif
    }

skip_check_filesize:

#endif


    if(strcmp(lpath,"/dev/null")==0) {
        Trace("write to black hole '%s'\n", lpath);
    }
    else
    if(ba) {
#ifdef __FreeBSD__
        chk = disable_disklabel_protect(ffd);
#endif
    }


#if 0
    iocheck(ffd);
#endif


    /***
     *** PASV for RETR
     ***/

start_store_pasv:
    fdprintf(cfd, "PASV\r\n");
try_store_pasv_read:
    nr = read(cfd, buf, BUFSIZ);
    if(nr < 0) {
        if(errno == EINTR)
            goto try_store_pasv_read;
    }
    if(strncasecmp(buf, "227", 3) != 0) {
        Error("fail FTP PASV for RETR <%d>", atoi(buf));

        if(atoi(buf)==226) {
            Trace("sleep and retry\n");
            fflush(stderr);
            fflush(stdout);
            sleep(3);
            goto start_store_pasv;
        }

        goto fail_data_conn;
    }
    {
        char data_addrstr[BUFSIZ];
        int data_port;
        struct sockaddr_in data_addr;
        int dd, dlen;

        chk = parse_ftp_peer(buf, data_addrstr, &data_port);

        Trace("data conn addr '%s' port %d\n", data_addrstr, data_port);

        chk = addr_bynumber(data_addrstr, data_port, &data_addr);

        dfd = TCP_connect_addr(&data_addr, data_port);

#if 0
        chk = getsockopt(dfd, SOL_SOCKET, SO_RCVBUF,
                (void*)&dd, &dlen);
        Trace("current RCVFBUF %d, %d\n", dd, dlen);
        if(chk<0) {
            Warn("fail getsockopt (%d)", errno);
        }

        Trace("set RCVBUF %d\n", rbsize);
        chk = setsockopt(dfd, SOL_SOCKET, SO_RCVBUF,
                (const void*)&rbsize, sizeof(rbsize));
        if(chk<0) {
            Warn("fail setsockopt (%d)", errno);
        }

        chk = getsockopt(dfd, SOL_SOCKET, SO_RCVBUF,
                (void*)&dd, &dlen);
        Trace("current RCVFBUF %d, %d\n", dd, dlen);
        if(chk<0) {
            Warn("fail getsockopt (%d)", errno);
        }
#endif

    }

    if(dfd < 0) {
        Error("cannot get data connection for FTP RETR");
        goto fail_data_conn;
    }
    Trace("data connection fd# %d\n", dfd);
    enlarge_sockbuf(dfd, SO_RCVBUF);


    /*
     * RETR
     */
    fdprintf(cfd, "RETR %s\r\n", rpath);
try_store_read:
    nr = read(cfd, buf, BUFSIZ);
    if(nr < 0) {
        if(errno == EINTR) {
            goto try_store_read;
        }
    }
    if(strncasecmp(buf, "150", 3) != 0) {
        Error("fail FTP RETR fd# %d, %s", cfd, buf);
        Error("maybe spell-miss about file or directory");
        goto skip_data_conn;
    }

#if 1
    MUTEX_LOCK(&info_lock);
    download_report = 1;
    gettimeofday(&download_start, NULL);
    download_end.tv_sec = 0;
    download_end.tv_usec = 0;
    MUTEX_UNLOCK(&info_lock);
#endif

    if(usez) {
        gzf = gzdopen(dfd, "rb");
        if(gzf==NULL) {
            Error("fail gzdopen() (%d)", errno);
        }
    }

    *hsaved = (off_t)0;

    cc = 0;
    mrt = 0;
    mwt = 0;
    rcn = rci = 0;
    rpcn = rpci = 0;
    wcn = wci = 0;
    while(1) {
        nr = 0;
        while(nr<blen) {
            gettimeofday(&mrst, NULL);
try_pread:
            if(usez) {
                pnr = gzread(gzf, &xbuf[nr], blen-nr);
            }
            else {
                pnr = read(dfd, &xbuf[nr], blen-nr);
            }
            if(pnr < 0) {
                Error("fail read() for RETR fd# %d, (%d)", dfd, errno);
                if(usez) {
                    gzclose(gzf);
                }
                else {
                    if(errno==EINTR) {
                        goto try_pread;
                    }
                }

                close(dfd);
                close(ffd);

                ret = 1;
                goto skip_data_conn;
            }
            gettimeofday(&mret, NULL);
            mrt += SUB_MSEC(mret, mrst);

            if(pnr==0) {
                goto inbuf_filled;
            }
            if(pnr%512==0) {
                rpcn++;
            }
            else {
                rpci++;
            }

            nr += pnr;
        }

inbuf_filled:
#if 0
        Trace("  nr %4d\n", nr);
#endif
        if(nr==0) {
            break;
        }


        len += nr;

        if(nr%512==0) {
            rcn++;
        }
        else {
            rci++;

            Trace(" irregular length nr %4d\n", nr);
            if(ba) {
                nr = ((nr + 512-1)/512)*512;
                Trace("  normalized nr %d\n", nr);
            }
        }

        cc++;
#if 0
        if(cc%RCREPINT==0) {
            time_t ccnow, ccelap;
            ccnow = time(NULL);
            ccelap = ccnow - download_start.tv_sec;
            Trace("%7d; cc %6d %6d; %lld/%lld; %d %d\n",
                ccnow,
                cc, ccelap,
                (long long)len, (long long)elen,
                mrt, mwt);
            Trace("rc  normal %7d, irregular %7d; total %d\n",
                rcn, rci, rcn+rci);
            Trace("rpc normal %7d, irregular %7d; total %d\n",
                rpcn, rpci, rcn+rci);
            Trace("wc  normal %7d, irregular %7d; total %d\n",
                wcn, wci, wcn+wci);
        }
#endif

        gettimeofday(&mwst, NULL);

try_write:
        nw = write(ffd, xbuf, nr);
        if(nw < 0) {
            if(errno == EINTR) {
                goto try_write;
            }
            Error("FTP-get fail write() disk fd# %d, (%d)", ffd, errno);
            if(usez) {
                gzclose(gzf);
            }
            close(dfd);
            close(ffd);
            goto cannot_write;
        }
        gettimeofday(&mwet, NULL);
        mwt += SUB_MSEC(mwet, mwst);

        if(nw!=nr) {
            wci++;
        }
        else {
            wcn++;
        }

#if 0
        MUTEX_LOCK(&info_lock);
        download_sum = (long long)len;
        MUTEX_UNLOCK(&info_lock);
#endif
#if 1
        MUTEX_LOCK(&info_lock);
        download_sum += (long long)nr;
        MUTEX_UNLOCK(&info_lock);
#endif

#if 0
        usleep(30000);
#endif
    }

    *hsaved = (off_t)len;

end:
#if 0
    Trace("cc %7d %d; %lld/%lld; %d %d\n",
        cc, time(NULL), (long long)len, (long long)elen,
        mrt, mwt);
    Trace("rc  normal %7d, irregular %7d; total %d\n",
        rcn, rci, rcn+rci);
    Trace("rpc normal %7d, irregular %7d; total %d\n",
        rpcn, rpci, rcn+rci);
    Trace("wc  normal %7d, irregular %7d; total %d\n",
        wcn, wci, wcn+wci);
#endif

    if(usez) {
        gzclose(gzf);
    }
    close(ffd);
    close(dfd);
    shutdown(dfd, 2);

skip_data_conn:
    fdprintf(cfd, "quit\r\n");
    close(cfd);
    shutdown(cfd, 2);

    MUTEX_LOCK(&info_lock);
    gettimeofday(&download_end, NULL);
    MUTEX_UNLOCK(&info_lock);

    goto out;


fail_data_conn:
    ret = -1;
    goto skip_data_conn;

cannot_write:
    ret = -1;
    goto skip_data_conn;

isdir:
nofile:
    Error("it is directory or not found");

    ret = 1;

out:
    Trace("lpath '%s'; got len %lld\n", lpath, (long long) len);
    return ret;
}



int
fix_port(char *pro)
{
    if(strcasecmp(pro, "http")==0) {
        return HTTP_PORTNO;
    }
    if(strcasecmp(pro, "ftp")==0) {
        return 21;
    }
    if(strcasecmp(pro, "tftp")==0) {
        return 69;
    }
#if 0
    if(strcasecmp(pro, "mft")==0) {
        return MFT_DOWN_DEFAULT_PORT;
    }
    if(strcasecmp(pro, "ffp")==0) {
        return FFP_PNO;
    }
#endif
    return -1;
}


/*
 * ParseURL - Parse URL and split to protocol, host, port and path.
 */
int
ParseURL(char *url, char *protocol, char *user, char *pass,
    char *host, char *port, char *path)
{
    register char *p,*q;
    register int c;
    int w;

    if(url==NULL)
        return 1;

    /* init */
    protocol[0] = '\0';
    user[0] = '\0';
    pass[0] = '\0';
    host[0] = '\0';
    port[0] = '\0';
    path[0] = '\0';

    /* Skip whites */
    p = url;
    while(*p&&(*p==' '||*p=='\t'||*p=='\r'||*p=='\n'))
        p++;
    url = p;


    /*
     * PROTOCOL
     */
    q = protocol;
    c = 0;
    while(c<BUFSIZ&&*p&&*p!=':') {
        *q++ = *p++;
        c++;
    }
    *q = '\0';

    /* No protocol */
    if(*p!=':') {
        protocol[0] = '\0';
        p = url;
    }
    else {
        p++;
    }

    /*
     * HOST and PORT
     */
    if(*p=='/' && *(p+1)=='/') {
        p++;
        p++;

        q = host;
        c = 0;
        while(c<BUFSIZ&&*p&&*p!='/') {
            *q++ = *p++;
            c++;
        }
        *q = '\0';

        /* check at-mark */
        q = host;
        c = 0;
        while(c<BUFSIZ&&*q&&*q!='@') {
            q++;
            c++;
        }
        /* user:pass @ host : port */
        if(*q=='@') {
            w = strlen(host);

            strncpy(user, host, q-host);
            user[q-host] = '\0';
            q++;

            memmove(host, q, w-(q-host)+1);

            /* check passwd */
            q = user;
            c = 0;
            while(c<BUFSIZ&&*q&&*q!=':') {
                q++;
                c++;
            }
            if(*q==':') {
                *q = '\0';
                q++;

                strcpy(pass, q);
            }
            else {
                pass[0] = '\0';
            }

        }

        /* check Port */
        q = host;
        c = 0;
        while(c<BUFSIZ&&*q&&*q!=':') {
            q++;
            c++;
        }
        if(*q==':') {
            *q = '\0';
            q++;

            strcpy(port, q);
        }
        /* Not port */
        else {
            port[0] = '\0';
        }


    }
    /* No host */
    else {
        host[0] = '\0';
        port[0] = '\0';
    }

    /*
     * PATH
     */
    q = path;
    c = 0;
    while(c<BUFSIZ&&*p) {
        *q++ = *p++;
        c++;
    }
    *q = '\0';

    return 0;
}

#ifndef HTTP_R_TOUT
#define HTTP_R_TOUT (3)
#endif

volatile int nret_http_read_timeout = S(HTTP_R_TOUT);

int
_http_get_m(int fd,
    char *aid, char *sid, char *sstat,
    char *epath, int *response_code, char *loc, char *hbuf, char *bbuf)
{
    char buf[BUFSIZ];
    char rbuf[BUFSIZ];
    int rlen;
    char line[BUFSIZ];
    register char *cpos;
    int blen;
    int etagcount;
    int *xetag;
    int nw;
    int chk;
    int r;
    char *b;
    int m;
    int n;
    char tmp[BUFSIZ];
    char c_type[BUFSIZ];
    int c_len;
    char etag[BUFSIZ];
    char *bstart;
    int ntry=5;

    Trace("_http_get_m: fd# %d\n", fd);
    fflush(stderr);

    sprintf(rbuf,
"GET %s HTTP/1.0\r\n\
User-Agent: ccftp/0.1\r\n\
X-NSS-Version: $Revision: 1.6 $\r\n\
X-NSS-Agent-ID: %s\r\n\
X-NSS-Session-ID: %s\r\n\
X-NSS-State: %s\r\n\
\r\n", epath, aid, sid, sstat);
    rlen = strlen(rbuf);


    dump_stringchunk("req", rbuf);
    fflush(stderr);

try_write:
    nw = write(fd, rbuf, rlen);
    if(nw < 0) {
        if(errno == EINTR)
            goto try_write;
        fprintf(stderr, "HTTP-get fail write (%d)\n", errno);
    }

    memset(buf, 0, BUFSIZ);

    *response_code = -1;
    *etag = '\0';
    m = 0;
    n = 0;
    while(n<BUFSIZ-1) {
#if 0
        chk = isreadable_via_select(fd, HTTP_R_TOUT, 0);
#endif
        chk = waitread_via_poll(fd, nret_http_read_timeout);
        if(chk==0) {
            Trace("timeout***\n");
            break;
        }
        
try_read:
        m = read(fd, &buf[n], BUFSIZ-n-1);
        Trace("m %d\n", m);
        if(m < 0) {
            if(errno == EINTR) {
                goto try_read;
            }
            Error("fail read() (%d)", errno);
            break;
        }
        if(m == 0) {
            break;
        }
        n += m;
        if(HTTP_find_header_term(buf, n) > 0) {
            break;
        }
    }


    cpos = buf;
    cpos = sgets(cpos, line, BUFSIZ);
    if(strlen(line) < 12) {
        Trace("1 response_header line #1 is too short (%dbytes).\n", n);
        return -1;
    }
    else {
        *response_code = HTTP_parse_response_code(line, strlen(line));
        Trace("2 response_header line #1 is enough.%d\n", *response_code);
    }

#if 0
    Trace("buf_line '%d', %s,buf  %s\n", n, line, buf);
#endif

    etagcount = 0;
    etag[0] = '\0';
    c_type[0] = '\0';
    c_len = -1;
    loc[0] = '\0';


    do {
        cpos = sgets(cpos, line, BUFSIZ);
        Chop_CRLF(line);
        dump_string("  h ", line);
        if(strncasecmp_first_all(line, 6, 'E', "ETag: ")) {
            strcpy(etag, &line[6]);
            etagcount ++;
        }
        else
        if(strncasecmp_first_all(line, 16, 'C', "Content-Length: ")) {
            strcpy(tmp, &line[16]);
            c_len = atoi(tmp);
        }
        else
        if(strncasecmp_first_all(line, 14, 'C', "Content-Type: ")) {
            strcpy(c_type, &line[14]);
        }
        else
        if(strncasecmp_first_all(line, 10, 'L', "Location: ")) {
            strcpy(loc, &line[10]);
        }
        else{
#if 0
            Trace("others '%s' \n",&line[0]);
#endif
        }
    } while(*line != '\0');

    bstart = cpos;
    memcpy(hbuf, buf, bstart-buf);
    if(n<=0) {
        Error("not contend ? n<=0");
        bbuf[0] = '\0';
    }
    else {
        memcpy(bbuf, bstart, n-(bstart-buf));
    }

    do {
        cpos = sgets(cpos, line, BUFSIZ);
        Chop_CRLF(line);
        dump_string("  b ", line);
    } while(*line != '\0');

    
    if(etagcount>0){
        Trace("find_etag '%s'\n",etag);
    }
    Trace("C-Type:   '%s'\n", c_type);
    Trace("C-Length: %d\n", c_len);
    Trace("Loc:      '%s'\n", loc);

#if 0
    dump_stringchunk("got head", hbuf);
    dump_stringchunk("got body", bbuf);
#endif

    return nw;
}



/*
*/

int
_http_get_d(int sfd,
    char *rpath, int *response_code, char *lpath, char *hbuf,
    off_t *hsaved)
{
    char buf[BUFSIZ];
    char rbuf[BUFSIZ];
    int rlen;
    char line[BUFSIZ];
    register char *cpos;
    int blen;
    int etagcount;
    int *xetag;
    int chk;
    int r;
    char *b;
    int m;
    int n;
    char tmp[BUFSIZ];
    char c_type[BUFSIZ];
    int c_len;
    char loc[BUFSIZ];
    char etag[BUFSIZ];
    char *bstart;
    int ntry=5;
    int ffd;
    int isdev;
    char bbuf[BUFSIZ];
    int have;
    int rest;
    int nr, nw;
    int snr, snw;
    struct stat sbuf;

    Trace("_htt_get_d: fd# %d\n", sfd);
    fflush(stderr);

#if 1
    ffd = open(lpath, O_WRONLY);
    if(ffd<0) {
        Warn("fail open() local for append 1 '%s' (%d)", lpath, errno);
        ffd = open(lpath, O_WRONLY|O_CREAT|O_TRUNC, 0644);
        if(ffd < 0) {
            Error("fail open() local for create 2 '%s' (%d)", lpath, errno);
#if 0
            goto nofile;
#endif
        goto out;
        }
    }

    chk = stat(lpath, &sbuf);
    if(chk) {
        Error("fail stat() local '%s' (%d)", lpath, errno);
#if 0
        goto nofile;
#endif
        goto out;
    }
    if(S_ISDIR(sbuf.st_mode)) {
        Error("local is directory '%s' (%d)", lpath);
#if 0
        goto isdir;
#endif
        goto out;
    }

    isdev = 0;
    if(is_devfile(lpath)) {
        Trace("local '%s' is device file (guessed)\n", lpath);
        if(blkcheck) {
            Trace("enter block align\n");
            isdev = 1;
        }
    }
#else
    ffd = assign_path(lpath, &isdev);
#endif

    if(ffd<0) {
        fprintf(stderr, "fail assign_path %d\n", ffd);
        return -1;
    }

#if 0
    sprintf(rbuf,
"GET %s HTTP/1.0\r\n\
User-Agent: ccftp/0.1\r\n\
X-NSS-Version: $Revision: 1.6 $\r\n\
X-NSS-Agent-ID: %s\r\n\
\r\n", rpath);
#endif

#if 0
    sprintf(rbuf,
"GET /%s HTTP/1.0\r\n\
User-Agent: Mozilla/4.0 (compatible; MSIE 6.0; Windows XP)\r\n\
\r\n", rpath);
#endif

#if 1
    sprintf(rbuf,
"GET /%s HTTP/1.0\r\n\
User-Agent: ccftp/0.1\r\n\
Accept: */*\r\n\
\r\n", rpath);
#endif

    rlen = strlen(rbuf);

    dump_stringchunk("req", rbuf);
    fflush(stderr);

try_write:
    nw = write(sfd, rbuf, rlen);
    if(nw < 0) {
        if(errno == EINTR)
            goto try_write;
        fprintf(stderr, "HTTP-get fail write (%d)\n", errno);
    }

    memset(buf, 0, BUFSIZ);

    *response_code = -1;
    *etag = '\0';
    m = 0;
    n = 0;
    while(n<BUFSIZ-1) {
#if 0
        chk = isreadable_via_select(sfd, HTTP_R_TOUT, 0);
#endif
        chk = waitread_via_poll(sfd, nret_http_read_timeout);
        if(chk==0) {
            Trace("timeout***\n");
            break;
        }
        
try_read:
        m = read(sfd, &buf[n], BUFSIZ-n-1);
        Trace("m %d\n", m);
        if(m < 0) {
            if(errno == EINTR) {
                goto try_read;
            }
            Error("fail read() (%d)", errno);
            break;
        }
        if(m == 0) {
            break;
        }
        n += m;
        if(HTTP_find_header_term(buf, n) > 0) {
            break;
        }
    }


    cpos = buf;
    cpos = sgets(cpos, line, BUFSIZ);
    if(strlen(line) < 12) {
        Trace("1 response_header line #1 is too short (%dbytes).\n", n);
        return -1;
    }
    else {
        *response_code = HTTP_parse_response_code(line, strlen(line));
        Trace("2 response_header line #1 is enough.%d\n", *response_code);
    }

#if 0
    Trace("buf_line '%d', %s,buf  %s\n", n, line, buf);
#endif

    etagcount = 0;
    etag[0] = '\0';
    c_type[0] = '\0';
    c_len = -1;
    loc[0] = '\0';


    do {
        cpos = sgets(cpos, line, BUFSIZ);
        Chop_CRLF(line);
        dump_string("  h ", line);
        if(strncasecmp_first_all(line, 6, 'E', "ETag: ")) {
            strcpy(etag, &line[6]);
            etagcount ++;
        }
        else
        if(strncasecmp_first_all(line, 16, 'C', "Content-Length: ")) {
            strcpy(tmp, &line[16]);
            c_len = atoi(tmp);
        }
        else
        if(strncasecmp_first_all(line, 14, 'C', "Content-Type: ")) {
            strcpy(c_type, &line[14]);
        }
        else
        if(strncasecmp_first_all(line, 10, 'L', "Location: ")) {
            strcpy(loc, &line[10]);
        }
        else{
#if 0
            Trace("others '%s' \n",&line[0]);
#endif
        }
    } while(*line != '\0');

#if 1
    if(c_len>=0) {
        MUTEX_LOCK(&info_lock);
        download_elen = (long long)c_len;
        MUTEX_UNLOCK(&info_lock);
    }
#endif

    bstart = cpos;
    memcpy(hbuf, buf, bstart-buf);
    if(n<=0) {
        Error("no content ? n<=0");
        bbuf[0] = '\0';
        goto empty;
    }


    memcpy(bbuf, bstart, n-(bstart-buf));
    have = n-(bstart-buf);
    rest = BUFSIZ - have;
    snr = have;
    snw = 0;

#if 1
    MUTEX_LOCK(&info_lock);
    download_sum += (long long)have;
    MUTEX_UNLOCK(&info_lock);
#endif

    while(1) {
try_bodyread:
        nr = read(sfd, &buf[have], rest);
        if(nr<0) {
            if(errno==EINTR)
                goto try_bodyread;
            fprintf(stderr, "fail read (%d)\n", errno);
            goto out;
        }
        if(nr==0) {
            fprintf(stderr, "reach to end\n");
            break;
        }
        snr += nr;
        have += nr;

#if 1
    MUTEX_LOCK(&info_lock);
    download_sum += (long long)nr;
    MUTEX_UNLOCK(&info_lock);
#endif

try_bodywrite:
        nw = write(ffd, bbuf, have);
        if(nw<0) {
            if(errno==EINTR)
                goto try_bodywrite;
            fprintf(stderr, "fail write (%d)\n", errno);
        }
        if(nw==0) {
        }
        if(nw!=have) {
            fprintf(stderr, "write mismattch have %d, write %d\n",
                have, nw);
        }

        snw += nw;
        have -= nw;
        rest = BUFSIZ - have;
    }

    close(ffd);


empty:

out:

    
    if(etagcount>0){
        Trace("find_etag '%s'\n",etag);
    }
    Trace("C-Type:   '%s'\n", c_type);
    Trace("C-Length: %d\n", c_len);
    Trace("Loc:      '%s'\n", loc);

#if 0
    *hsaved = c_len;
#endif
    *hsaved = (off_t)snr;

#if 0
    dump_stringchunk("got head", hbuf);
    dump_stringchunk("got body", bbuf);
#endif

    return nw;
}


int
HTTP_short_get(char *url, char *aid, char *sid, char *smsg,
 int *rcode, char *loc, char *hbuf, char *bbuf)
{
    int sret;
    int sfd;
    int chk;
    char protocol[BUFSIZ];
    char user[BUFSIZ];
    char pass[BUFSIZ];
    char host[BUFSIZ];
    char port[BUFSIZ];
    int portno;
    char path[BUFSIZ];
    char tags[BUFSIZ];


    sret = -1;

    chk = ParseURL(url, protocol, user, pass, host, port, path);

    portno = atoi(port);

    fprintf(stderr, "URL      '%s'\n", url);
#if 0
    fprintf(stderr, "protocol '%s'\n", protocol);
    fprintf(stderr, "user     '%s'\n", user);
    fprintf(stderr, "pass     '%s'\n", pass);
    fprintf(stderr, "host     '%s'\n", host);
    fprintf(stderr, "port     '%s' %d\n", port, portno);
    fprintf(stderr, "path     '%s'\n", path);
#endif

    if(portno<=0) {
        portno = fix_port(protocol);
        fprintf(stderr, "set port to %d\n", portno);
    }

    sfd = TCP_connect(host, portno);
    if(sfd<0) {
        fprintf(stderr, "fail connect to '%s':%d\n",
            host, portno);
        goto out;
    }

    fprintf(stderr, "HTTP_GET fd# %d\n", sfd);
    fflush(stderr);
    sret = _http_get_m(sfd, aid, sid, smsg, path, rcode, loc, hbuf, bbuf);

    fprintf(stderr, "rcode %d\n", *rcode);

    shutdown(sfd, 2);
    close(sfd);

    sret = 0;

out:
    return sret;
}



int
HTTP_get(char *host, int portno, char *rpath, char *lpath,
    off_t *hsaved)
{
    int sret;
    int sfd;
    int chk;
#if 0
    char protocol[BUFSIZ];
    char user[BUFSIZ];
    char pass[BUFSIZ];
    char host[BUFSIZ];
    char port[BUFSIZ];
    int portno;
    char path[BUFSIZ];
    char tags[BUFSIZ];
#endif
    int rcode;
    char hbuf[BUFSIZ];

    sret = -1;

    sfd = TCP_connect(host, portno);
    if(sfd<0) {
        fprintf(stderr, "fail connect to '%s':%d\n",
            host, portno);
        goto out;
    }

    fprintf(stderr, "HTTP_GET fd# %d\n", sfd);
    fflush(stderr);
    sret = _http_get_d(sfd, rpath, &rcode, lpath, hbuf, hsaved);

    fprintf(stderr, "rcode %d\n", rcode);

    shutdown(sfd, 2);
    close(sfd);

    sret = 0;

out:
    return sret;
}



int
download(char *src, char *dst, off_t *hsaved)
{
    char spro[BUFSIZ];
    char suser[BUFSIZ];
    char spass[BUFSIZ];
    char shost[BUFSIZ];
    char sport[BUFSIZ];
    int  sportno;
    char spath[BUFSIZ];
    char dpro[BUFSIZ];
    char duser[BUFSIZ];
    char dpass[BUFSIZ];
    char dhost[BUFSIZ];
    char dport[BUFSIZ];
    int  dportno;
    char dpath[BUFSIZ];
    int cfd;
    int ck;
    int rcode;

    Trace("download: '%s' '%s'\n", src, dst);

    ParseURL(src, spro, suser, spass, shost, sport, spath);
    sportno = atoi(sport);
    if(sportno<=0) {
        sportno = fix_port(spro);
        Trace("set src port to %d\n", sportno);
    }
    Trace("source ---\n");
    Trace("URL      '%s'\n", src);
    Trace("protocol '%s'\n", spro);
    Trace("user     '%s'\n", suser);
    Trace("pass     '%s'\n", spass);
    Trace("host     '%s'\n", shost);
    Trace("port     '%s' %d\n", sport, sportno);
    Trace("path     '%s'\n", spath);


    ParseURL(dst, dpro, duser, dpass, dhost, dport, dpath);
    dportno = atoi(dport);
    if(dportno<=0) {
        dportno = fix_port(dpro);
        Trace("set dst port to %d\n", dportno);
    }
    Trace("destination ---\n");
    Trace("URL      '%s'\n", dst);
    Trace("protocol '%s'\n", dpro);
    Trace("user     '%s'\n", duser);
    Trace("pass     '%s'\n", dpass);
    Trace("host     '%s'\n", dhost);
    Trace("port     '%s' %d\n", dport, dportno);
    Trace("path     '%s'\n", dpath);

    if(strcasecmp(spro, "ftp")==0) {
        cfd = TCP_connect(shost, sportno);
        if(cfd<0) {
            if(use_syslog) {
                syslog(LOG_ERR, "fail connect via FTP GET for '%s'", src);
            }
            return -1;
        }
        /* XXX */
        ck = FTP_get(cfd, suser, spass, spath+1, dpath, hsaved);
        Trace("FTP_get returns %d\n", ck);
        if(ck) {
            if(use_syslog) {
                syslog(LOG_ERR, "fail download via FTP GET for '%s'", src);
            }
            return -1;
        }
    }
    else
    if(strcasecmp(spro, "http")==0) {
        ck = HTTP_get(shost, sportno, spath+1, dpath, hsaved);
        Trace("HTTP_get returns %d\n", ck);
    }
#if 0
    else
    if(strcasecmp(spro, "ffp")==0) {
        extern volatile int ffpget_trace;
        extern volatile int xmft_trace;

        ffpget_trace = 31;
        xmft_trace = 31;
        ck = ffp_get(shost, sportno, spath+1, dpath);
        Trace("FFP_get returns %d\n", ck);
        if(ck) {
            return ck;
        }
    }
#endif
    else {
        Error("unknown porotocol");
    }

    return 0;
}

int
upload(char *src, char *dst)
{
    char spro[BUFSIZ];
    char suser[BUFSIZ];
    char spass[BUFSIZ];
    char shost[BUFSIZ];
    char sport[BUFSIZ];
    int  sportno;
    char spath[BUFSIZ];
    char dpro[BUFSIZ];
    char duser[BUFSIZ];
    char dpass[BUFSIZ];
    char dhost[BUFSIZ];
    char dport[BUFSIZ];
    int  dportno;
    char dpath[BUFSIZ];
    int cfd;
    int ck;

    Trace("upload: '%s' '%s'\n", src, dst);
    
    ParseURL(src, spro, suser, spass, shost, sport, spath);
    sportno = atoi(sport);
    if(sportno<=0) {
        sportno = fix_port(spro);
        fprintf(stderr, "set src port to %d\n", sportno);
    }
    Trace("URL      '%s'\n", src);
    Trace("protocol '%s'\n", spro);
    Trace("user     '%s'\n", suser);
    Trace("pass     '%s'\n", spass);
    Trace("host     '%s'\n", shost);
    Trace("port     '%s' %d\n", sport, sportno);
    Trace("path     '%s'\n", spath);

    ParseURL(dst, dpro, duser, dpass, dhost, dport, dpath);
    dportno = atoi(dport);
    if(dportno<=0) {
        dportno = fix_port(dpro);
        Trace("set dst port to %d\n", dportno);
    }
    Trace("URL      '%s'\n", dst);
    Trace("protocol '%s'\n", dpro);
    Trace("user     '%s'\n", duser);
    Trace("pass     '%s'\n", dpass);
    Trace("host     '%s'\n", dhost);
    Trace("port     '%s' %d\n", dport, dportno);
    Trace("path     '%s'\n", dpath);

    if(strcasecmp(dpro, "ftp")==0) {
        cfd = TCP_connect(dhost, dportno);
        /* XXX */
        ck = FTP_put(cfd, duser, dpass, dpath+1, spath);
        Trace("FTP_put returns %d\n", ck);
        if(ck) {
            if(use_syslog) {
                syslog(LOG_ERR, "fail upload via FTP PUT for '%s'", src);
            }
            return -1;
        }
    }
#if 0
    else
    if(strcasecmp(dpro, "http")==0) {
        cfd = TCP_connect(dhost, dportno);
        ck = HTTP_post(cfd, dpath, duser, dpass);
        Trace("HTTP_post returns %d\n", ck);
    }
#endif
    else {
    }
    
    return 0;
}



